/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Copyright (C) 2002 Motorola Semiconductors HK Ltd
 *
 */


/******************************************************************************
 *
 * Copyright (C) 2001, Motorola All Rights Reserved
 *
 * File Name:   mmcsd.c
 *
 * Progammers:  Yiwei Zhao
 *
 * Date of Creations:   30  DEC,2001
 *
 * Synopsis:
 *
 * Modification History:
 * 30 DEC, 2001
 * 24,Dec,2003 Karen Kang
 *****************************************************************************/

#define TESTING_PAGE (1024*1024*40/512)
#define MMCSD_TEST_PAGE_NUM 8

#include <system_def.h>
#include <string.h>
#include <malloc.h>
#include <errno.h>
#include <mx1_reg.h>
#include <drvsupport.h>
#include <stdio.h>
#include "mmcsd_def.h"
#include "mmcsd_main.h"        /* local definitions */

#ifdef MMCSD_WITH_RTEMS

/*RTEMS specific hacks*/
#define copy_to_user memcpy
#define copy_from_user memcpy
#define __virt_to_phys(x) (x)

#endif /*MMCSD_WITH_RTEMS*/

/*These symbols in <linux/blk.h> that must be defined in advance.*/
#define DEVICE_NR(device) MINOR(device)
#define DEVICE_NAME "mmc"			/* name for messaging */

/*
 * if partition is considered then HAVE_BLKPG_H should be defined.
 */

/*
 * variables will be initialized in init function.
 */
uint8_t			g_mmcsd_num = 0;
MMCSD_Dev 		*g_mmcsd_devices[MMCSD_DEVICE_NUM];
MMCSD_Lock		g_selected_lock[MMCSD_DEVICE_NUM];
int 			g_mmcsd_major;
uint32_t		g_mmcsd_irq_count;

/* Function declarations */

int mmcsd_read_device(	MMCSD_Dev *device, char * buffer,
		uint32_t sector, uint32_t nr_sector);

int mmcsd_write_device( MMCSD_Dev *device, char * buffer,
		 uint32_t sector, uint32_t nr_sector);

static int mmcsd_sector_erase(MMCSD_Dev *device, MMCSD_Blks *blks);

MMCSD_STATUS mmcsd_ReadBlock(MMCSD_Dev *device,uint32_t start_addr,uint16_t nsects, uint8_t *buff);

MMCSD_STATUS mmcsd_ReadSingleBlock(MMCSD_Dev *device,uint32_t addr,uint8_t * buff);

#ifdef MMCSD_READ_DMA
MMCSD_STATUS mmcsd_DMA_ReadSingleBlock(MMCSD_Dev *device,uint32_t addr,uint8_t * buff);
#endif /*MMCSD_READ_DMA*/

#if defined(MMCSD_READ_DMA) || defined(MMCSD_WRITE_DMA)
int mmcsd_DmaStartTransfer(MMCSD_Dev *device,uint8_t * buff,uint32_t counter, uint8_t mode,uint8_t is_4_bit);
#endif

void mmcsd_dma_handler(void);

//void mmcsd_tasklet(void);

int mmcsd_ioctl_device(MMCSD_Dev *device, unsigned command, unsigned long  argument);

static int mmcsd_write_protect(MMCSD_Dev *device,uint32_t addr, char is_set);

static int mmcsd_passwd_cmd(MMCSD_Dev * device, MMCSD_PWD *pwd);

static int mmcsd_is_writeprotection(MMCSD_Dev *device);

#ifdef MMCSD_WITH_RTEMS
rtems_isr mmcsd_irq_handler(rtems_irq_hdl_param data);
#else /*MMCSD_WITH_RTEMS*/
void mmcsd_irq_handler(int irq,void * data, struct pt_regs * pt);
#endif /*MMCSD_WITH_RTEMS*/

int mmcsd_getMBR(MMCSD_Dev * device);

static int mmcsd_check_status(MMCSD_Dev *device,uint32_t *pstate);

static void mmcsd_dma_isr(void * arg);

static void mmcsd_dma_err_isr(int error_type);

MMCSD_STATUS mmcsd_write_blks(MMCSD_Dev *device);

#ifdef MMCSD_WRITE_DMA
MMCSD_STATUS mmcsd_dma_write_blks(MMCSD_Dev *device);
#endif /*MMCSD_READ_DMA*/

void mmcsd_free_buffer(MMCSD_Dev *device);

void _mmcsd_16_big_2_little(uint8_t *buf, uint32_t size);

void _mmcsd_cp_16_little_2_big(uint8_t *src,uint8_t *target, int size);

void mmcsd_delay(uint32_t delay);

#ifdef MMCSD_WITH_PM
static int handle_pm_event(struct pm_dev *dev, pm_request_t rqst, void *data);
#endif /*MMCSD_WITH_PM*/

/* Set up our tasklets. */

extern unsigned char  gSDWriteProtect;

static char g_socket1_info;
#define SD_SOCKET_NO_CARD		0x01
#define SD_SOCKET_CARD			0x02

#ifndef MMCSD_WITH_RTEMS
inline int mmcsd_device_select_init(MMCSD_Dev *device)
{
	init_waitqueue_head(&(device->select_wait));
	return 0;
}

inline long mmcsd_device_select_wait(MMCSD_Dev *device, long timeout)
{
	return 10*interruptible_sleep_on_timeout(&(device->select_wait),timeout/10);
}

inline int mmcsd_device_select_wake(MMCSD_Dev *device)
{
	wake_up_interruptible(&(device->select_wait));
	return 0;
}
#endif /*MMCSD_WITH_RTEMS*/

/*
 ********************************************************************************
 * Description:
 *       Card detect function
 *
 * History:
 *       03-Feb-04       Aaron   Add Card detection for MX1 silicon 2.0 (Use PB17)
 ********************************************************************************
 */
#ifndef MMCSD_WITH_RTEMS
void mmcsd_socket1_irq_handler(int irq, void *dev_id, struct pt_regs * regs)
{
	uint32_t status,trigger;
#ifdef CONFIG_ARCH_MX1ADS
	status = _reg_PTB_ISR;
	trigger = _reg_PTB_ICR2;
	if((status & 0x00100000)==0x00100000) //get PortB20 int
	{
		if((trigger & 0x00000300)==0x00000200) //it is positive level
		{
			g_socket1_info = SD_SOCKET_NO_CARD;
			_reg_PTB_ICR2 |= 0x00000300; //set it to be negative level trigger detect card withdraw
			TRACE(("A card is withdrawn from socket1 \n");
		}
		else if((trigger & 0x00000300)==0x00000300) //it is negative level
		{
			g_socket1_info = SD_SOCKET_CARD;
			_reg_PTB_ICR2 &= ~0x00000100;  // config as positive level trigger10
			_reg_PTB_ICR2 |= 0x00000200; ;//set it to be positive level trigger detect card insert
			TRACE(("A card is inserted in socket1 \n");
		}
	}
#endif
#ifdef CONFIG_ARCH_MX2ADS
	status = _reg_GPIO_ISR(GPIOD);
	trigger = _reg_GPIO_ICR2(GPIOD);
	if((status & 0x02000000)==0x02000000) //get PortB16 int
	{
		if((trigger & 0x000C0000)==0x00080000)//it is positive level now
		{
			g_socket1_info = SD_SOCKET_NO_CARD;
			_reg_GPIO_ICR2(GPIOD) |= 0x000C0000; //set it to be negative level trigger detect card withdraw
			TRACE(("A card is withdrawn from socket1 \n");
		}
		if((trigger & 0x000C0000)==0x000C0000)//it is negative level now
		{
			g_socket1_info = SD_SOCKET_CARD;
			_reg_GPIO_ICR2(GPIOD) &= ~0x00040000;  // config as positive level trigger10
			_reg_GPIO_ICR2(GPIOD) |= 0x00080000;//set it to be positive level trigger detect card insert
			TRACE(("A card is inserted in socket1 \n");
		}
	}
#endif
	_mmcsd_socket1_clear_irq();
}
#endif /*MMCSD_WITH_RTEMS*/

#ifndef MMCSD_WITH_RTEMS
static int mmcsd_proc_output (char *buf)
{
	char *p;
	unsigned long flags;
	save_flags(flags);
	cli();
	p = buf;
	//Add to proc to indicate which card in which socket

	if(g_socket1_info == SD_SOCKET_NO_CARD)
	{
		p += sprintf(p, "socket1 : nocard \n");
		//		p += sprintf(p, "socket2 : nocard \n");
	}
	else if(g_socket1_info == SD_SOCKET_CARD)
	{
		p += sprintf(p, "socket1 : card \n");
	}
	restore_flags(flags);
	return p - buf;

}

static int mmcsd_read_proc(char *page, char **start, off_t off,
		int count, int *eof, void *data)
{

	int len = mmcsd_proc_output(page);
	if (len <= (off+count))
		*eof = 1;
	*start = page + off;
	len -= off;
	if (len>count)
		len = count;
	if (len<0)
		len = 0;
	return len;
}
#endif /*MMCSD_WITH_RTEMS*/

/******************************************************************************
 * Function Name: mmcsd_init
 *
 * Input:		None.
 * Value Returned:	0:	success;
 *			others 	failure;
 *
 * Description: This is the initialization routine for the driver. And this function
 *		will be called while the module being installed.
 *
 * Modification History:
 * 	30 DEC,2001, Initial version.  			Yiwei Zhao
 *****************************************************************************/
int mmcsd_init(void)
{

	int ret=ENODEV;
	int i;
	MMCSD_Dev *device;
	printk("MMC device driver version 0.4.0\n");
	TRACE("Begin initialize mmcsd module\n");

	/*******************************************************
	 * Allocate space for each devices.
	 *******************************************************/

#ifdef CONFIG_ARCH_MX2ADS
	g_mmcsd_num = MMCSD_DEVICE_NUM;
#else
	g_mmcsd_num = 1;
#endif

	for (i=0; i < g_mmcsd_num; i++)
	{
		device = malloc(sizeof(MMCSD_Dev));
		if (!device)
			goto fail_init;

		memset(device,0,(sizeof(MMCSD_Dev)));

		device->port = i;
#ifdef CONFIG_ARCH_MX2ADS
		if(i==1){
			device->base_addr=MX1_MMCSD_BASE_2;
			device->irq_number=MMCSD2_IRQ;
		}else
#endif
		{
			device->base_addr=MX1_MMCSD_BASE;
			device->irq_number=MMCSD_IRQ;
		}
		mmcsd_device_lock_init(device);
		mmcsd_device_select_init(device);
		mmcsd_device_sem_init(device);
		device->dma_channel = 0xFFFFFFFF;

#ifdef MMCSD_WITH_RTEMS
		ret = drvsup_request_irq(device->irq_number, mmcsd_irq_handler,
					0, "mmc", device);
#else /*MMCSD_WITH_RTEMS*/
		ret = request_irq(device->irq_number, mmcsd_irq_handler,
				SA_INTERRUPT, "mmc", device);
#endif /*MMCSD_WITH_RTEMS*/

		if(ret)
		{
			TRACE("request irq error\n");
			goto fail_init;
		}


		g_mmcsd_devices[i]=device;

		g_selected_lock[i].device = NULL;
	}

	/******************************************************
	 * Regist Interrupt Service Routine
	 ******************************************************/

#ifndef MMCSD_WITH_RTEMS
	/*****************************************************
	 *setting for auto-mount/umount
	 ******************************************************/
	g_socket1_info = SD_SOCKET_NO_CARD;
	ret = request_irq(MMCSD_SOCKET1_INT,
			mmcsd_socket1_irq_handler,
			SA_INTERRUPT,
			"mmc",
			"mmc");
#ifdef CONFIG_ARCH_MX2ADS
	ret = request_irq(MMCSD_SOCKET1_INT,
			mmcsd_socket1_irq_handler,
			SA_INTERRUPT|SA_SHIRQ,
			"mmc",
			"mmc");
#endif
	if(ret)
	{
		TRACE("request irq error\n");
		goto fail_init;
	}
	_mmcsd_socket1_irq_set();
	create_proc_read_entry ("mmc", 0, 0, mmcsd_read_proc, NULL);
#endif /*MMCSD_WITH_RTEMS*/
	/******************************************************
	 * Regist PM device.
	 ******************************************************/

#ifdef MMCSD_WITH_PM
	 pm_register(PM_SYS_DEV, PM_SYS_UNKNOWN, handle_pm_event);
#endif /*MMCSD_WITH_PM*/

	/******************************************************
	 * Pin Configuration
	 *
	 *****************************************************/

	_MMCSD_64M_MMC_PinConfig();

	/******************************************************
	 *  LOOP FOR THE MAXIMUM NUMBER OF CARD SLOTS
	 ******************************************************/
	for ( i = 0 ; i < g_mmcsd_num ; i++ )
	{
		g_mmcsd_devices[i]->card_state= MMCSD_NOTREADY;
		g_mmcsd_devices[i]->card_rca = 1;
	}

	return 0;

	/******************************************************
	 * Enable SDHC module
	 ******************************************************/

fail_init:
	for ( i = 0 ; i < g_mmcsd_num ; i++ )
	{
		device = g_mmcsd_devices[i];
		g_mmcsd_devices[i] = NULL;

#ifndef MMCSD_WITH_RTEMS
		if(device->irq_number)
			free_irq(device->irq_number, device);
#else /*MMCSD_WITH_RTEMS*/
#endif /*MMCSD_WITH_RTEMS*/

		if(device)
			free(device);

	}
#ifdef 	CONFIG_ARCH_MX2ADS
	free_irq(MMCSD_SOCKET1_INT,"mmc");
#endif
	return ret;
}

/******************************************************************************
 * Function Name: mmcsd_cleanup
 *
 * Input:		None.
 * Value Returned:	0:	success;
 *			others 	failure;
 *
 * Description:
 *	this routine will clean up mmcsd module.
 *
 * Modification History:
 * 	30 DEC,2001, Initial version.  			Yiwei Zhao
 *****************************************************************************/
int mmcsd_cleanup(void)
{
	int i;

#ifdef MMCSD_WITH_PM
	pm_unregister_all(handle_pm_event);
#endif /*MMCSD_WITH_PM*/
	for ( i = 0 ; i < g_mmcsd_num ; i++ )
	{
		MMCSD_Dev *device = g_mmcsd_devices[i];
		g_mmcsd_devices[i] = NULL;

#ifndef MMCSD_WITH_RTEMS
		if(device->irq_number)
			free_irq(device->irq_number, device);
#else /*MMCSD_WITH_RTEMS*/
#endif /*MMCSD_WITH_RTEMS*/

		if(device)
			free(device);
	}

#ifdef 	CONFIG_ARCH_MX2ADS
	free_irq(MMCSD_SOCKET1_INT,"mmc");
	remove_proc_entry("mmc", NULL);
#endif
	return 0;

}

/******************************************************************************
 * Function Name: mmcsd_revalidate
 *
 * Input:		None.
 * Value Returned:	0:	success;
 *			others 	failure;
 *
 * Description: This routine will revalidate device.
 *
 * Modification History:
 * 	30 DEC,2001, Initial version.  			Yiwei Zhao
 *****************************************************************************/
int  mmcsd_revalidate(MMCSD_Dev * device)
{
	return 0;
}
/******************************************************************************
 * Function Name: mmcsd_open_device
 *
 * Input:		None.
 * Value Returned:	0:	success;
 *			others 	failure;
 *
 * Description: This routine will initilialize a device.
 *
 * Modification History:
 * 	30 DEC,2001, Initial version.  			Yiwei Zhao
 *****************************************************************************/
int mmcsd_open_device(MMCSD_Dev *device)
{
	int num,i,ret;
	int	channel=-1;

	TRACE("mmcsd_open is called\n");
	/*********************************************************************
	 * Check the device numer
	 *********************************************************************/
	num = device->port;
	gSDWriteProtect = 0;

	if(num > g_mmcsd_num)
	{
		return -ENODEV;
	}

#ifdef CONFIG_ARCH_MX2ADS
	if(device->port == 0)
		_reg_CRM_PCCR0 |= 0x0200;//enable sdhc1 clk
	else if(device->port == 1)
		_reg_CRM_PCCR0 |= 0x0400;//enable sdhc2 clk

#endif

	/*****************************************************************
	 * Initialize the device
	 *****************************************************************/
	mmcsd_device_lock(device);
	mmcsd_device_sem_obtain(device);

	ret = 0;
	if(device->usage == 0)
	{
		/******************************************************
		 * Initialize Cards
		 ******************************************************/
		_MMCSD_64M_MMC_PinConfig();
		_MMCSD_InitSDHC(device);
		ret = _MMCSD_InitMMC(device);
		if(ret)
		{
			FAILED("_MMC_Init error\n");
			goto fail_malloc;
			//			return -1;
		}
		ret = _MMCSD_GetInfo(device);	// Implement
		if(ret)
		{
			FAILED("_MMCSD_GetInfo error!\n");
			mmcsd_device_unlock(device);
			mmcsd_device_sem_release(device);
			goto fail_malloc;
			//			return -1;
		}
		ret = _MMCSD_SetSectLen(device);	// Implement
		if(ret)
		{
			FAILED(" _MMCSD_SetSectLen error!\n");
			mmcsd_device_unlock(device);
			mmcsd_device_sem_release(device);
			goto fail_malloc;
			//			return -1;
		}
		device->blksize = MMCSD_MINIMUM_BLK_SIZE;

		ret = _MMCSD_SelectDeselectCard(device, device->card_rca);
		if(ret)
		{
			FAILED("Select error\n");
			device->card_state = MMCSD_NOTREADY;
			goto fail_malloc;
		}
		/***********************************************************
		 * Initialize DMA
		 ***********************************************************/
#if defined(MMCSD_READ_DMA) || defined(MMCSD_WRITE_DMA)
#ifdef CONFIG_ARCH_MX1ADS
		for(channel = 2; channel < 11; channel++)
		{
			if((ret = request_dma(channel, "mmc"))==0)
			{
				//TRACE("Get Channel %ld\n",channel);
				break;
			}
		}
		if(channel >10) goto fail_malloc;

		device->dma_channel = channel;
		request_dma_intr(channel,(void *)mmcsd_dma_isr,(void *)mmcsd_dma_err_isr);
#endif
#ifdef CONFIG_ARCH_MX2ADS
		if(device->port == 1)
		{
			mx_request_dma(&channel,"mmc1");
			(device->dma).callback = (void *)mmcsd_dma_isr_1;
		}
		else
		{
			mx_request_dma(&channel,"mmc");
			(device->dma).callback = (void *)mmcsd_dma_isr;
		}
		device->dma_channel = channel;
#endif
#endif /* defined(MMCSD_READ_DMA) || defined(MMCSD_WRITE_DMA) */

		/***********************************************************
		 * Initialize buff_pool for write operations.
		 **********************************************************/

		device->buff_pool[0]=malloc(sizeof(uint8_t)*(device->blksize)*MMCSD_MAX_MULTI_BLK);
#ifdef MMCSD_DMA_ENDIAN_ERR
		device->tmp_pool[0]=malloc(sizeof(uint8_t)*(device->blksize)*MMCSD_MAX_MULTI_BLK);
#endif

		if(device->buff_pool[0]==NULL)
		{
			FAILED("Allocate Buffer failed\n");
			goto fail_malloc;
		}
		for(i = 0; i< MMCSD_MAX_MULTI_BLK;i++)
		{
			device->buff_pool[i] = device->buff_pool[0]+(device->blksize)*i;
#ifdef MMCSD_DMA_ENDIAN_ERR
			device->tmp_pool[i] = device->tmp_pool[0]+(device->blksize)*i;
#endif
		}
		device->cur_sec = 0xFFFFFFFF;
		device->cur_num = 0;
		device->busy = 0;

		device->bit_width = 0;
		_MMCSD_SetClk(device,device->max_dtr);

		/***********************************************************
		 * Get MBR
		 ***********************************************************/

#if 0
		/*FIXME: There is no reason to play partition table games in this layer*/
		ret = mmcsd_getMBR(device);
		if(ret)
		{
			FAILED("mmcsd_getMBR error!\n");
			goto fail_malloc;
		}
#endif
		/***********************************************************
		 * Detect wether it's read only
		 ***********************************************************/

		ret = 0;
	}
	(device->usage)++;
	mmcsd_device_unlock(device);
	mmcsd_device_sem_release(device);
	return ret;

fail_malloc:
#if defined(MMCSD_READ_DMA) || defined(MMCSD_WRITE_DMA)
#ifdef CONFIG_ARCH_MX1ADS
	if(device->dma_channel!= 0xFFFFFFFF)
		free_dma(device->dma_channel);
#endif
#ifdef CONFIG_ARCH_MX2ADS
	if(device->dma_channel!= 0xFFFFFFFF)
		mx_free_dma(device->dma_channel);
#endif
#endif /* defined(MMCSD_READ_DMA) || defined(MMCSD_WRITE_DMA) */

	if(device->buff_pool[0]!=NULL)
		free(device->buff_pool[0]);

	mmcsd_device_unlock(device);
	mmcsd_device_sem_release(device);
	return MMCSD_ERROR;

}

/******************************************************************************
 * Function Name: mmcsd_release_device
 *
 * Input:		None.
 * Value Returned:	0:	success;
 *			others 	failure;
 *
 * Description: This routine will release the device.
 *
 * Modification History:
 * 	30 DEC,2001, Initial version.  			Yiwei Zhao
 *****************************************************************************/

int mmcsd_release_device(MMCSD_Dev * device)
{
	uint32_t	state;
	MMCSD_STATUS ret;

	TRACE("mmcsd_release is called\n");

	/*****************************************************************
	 * Initialize the device
	 *****************************************************************/
	mmcsd_device_lock(device);
	mmcsd_device_sem_obtain(device);

	(device->usage)--;

	if(device->usage == 0)
	{
		/*fsync_dev(inode->i_rdev);*/

#ifdef MMCSD_WRITE_DMA
		mmcsd_dma_write_blks(device);
#else
		mmcsd_write_blks(device);
#endif
		state = MMCSD_TRANSFER;
		mmcsd_check_status(device,&state);

		ret = _MMCSD_SelectDeselectCard(device, 0);
		if(ret)
		{
			TRACE("Select error\n");
			device->card_state = MMCSD_NOTREADY;
		}

		//		_MMCSD_StopClk(device);
		//		_MMCSD_StartStopClk(device,1);
		g_selected_lock[device->port].device = NULL;

#if defined(MMCSD_READ_DMA) || defined(MMCSD_WRITE_DMA)
#ifdef CONFIG_ARCH_MX1ADS
		if(device->dma_channel!= 0xFFFFFFFF)
			free_dma(device->dma_channel);
#endif
#ifdef CONFIG_ARCH_MX2ADS
		if(device->dma_channel!= 0xFFFFFFFF)
			mx_free_dma(device->dma_channel);
#endif
#endif /* defined(MMCSD_READ_DMA) || defined(MMCSD_WRITE_DMA) */

		if(device->buff_pool[0]!=NULL)
			free(device->buff_pool[0]);
#ifdef MMCSD_DMA_ENDIAN_ERR
		if(device->buff_pool[0]!=NULL)
			free(device->buff_pool[0]);
#endif
	}

	mmcsd_device_unlock(device);
	mmcsd_device_sem_release(device);

	return 0;
}

/******************************************************************************
 * Function Name: mmcsd_check_status
 *
 * Input:
 * 		device: the device read from
 * 		buffer:	memory area for I/O transfer.
 * 		sector: first sector number;
 * 		nr_sector: number of the sectos
 *
 * Value Returned:
 * 		0:	success.
 * 		1:	failure.
 *
 * Description:
 * 	this routine will read indicated sectors into memory
 *
 * Modification History:
 * 	30 DEC,2001, Initial version.  			Yiwei Zhao
 *****************************************************************************/

int mmcsd_check_status(MMCSD_Dev *device,uint32_t *pstate)
{
	MMCSD_STATUS 	ret = MMCSD_OK;
	uint32_t		state;
	int	     	i;

	state = *pstate;

	device->cur_cmd=MMCSD_CMD13;

	i = 0xFFFFFFFF;
	while(i--)
	{
		//TRACE("SendStatus... 0x%lx\n",mmcsd_reg_rd(device, MX1_MMCSD_STATUS_o));
		ret = _MMCSD_SendStatus(device,device->card_rca,pstate);
		if(ret)	// the command failed.
		{
			mmcsd_delay(0xFF);
			//TRACE("Card Status = 0x%08X\n", *pstate);
		}
		else
		{
			if(((*pstate)&MMCSD_STATUS_STATE) != state)
			{
				mmcsd_delay(0xFF);
			}
			else
			{
				if(state==MMCSD_TRANSFER)
				{
					if(((*pstate)&0x100)==0)
						continue;
				}
				break;
			}
			TRACE("state = 0x%lx,request is 0x%lx\n",(unsigned long)*pstate,(unsigned long)state);
		}
		//TRACE("try again..(SendStatus... 0x%lx) %d\n",mmcsd_reg_rd(device, MX1_MMCSD_STATUS_o),i);

	}
	//TRACE("return with %d\n",ret);
	return ret;
}

/******************************************************************************
 * Function Name: mmcsd_read_device
 *
 * Input:
 * 		device: the device read from
 * 		buffer:	memory area for I/O transfer.
 * 		sector: first sector number;
 * 		nr_sector: number of the sectos
 *
 * Value Returned:
 * 		0:	success.
 * 		1:	failure.
 *
 * Description:
 * 	this routine will read indicated sectors into memory
 *
 * Modification History:
 * 	30 DEC,2001, Initial version.  			Yiwei Zhao
 *****************************************************************************/

int mmcsd_read_device(	MMCSD_Dev *device, char * buffer,
		uint32_t sector, uint32_t nr_sector)
{
	int ret;
	uint32_t	start_addr,state;

	mmcsd_device_sem_obtain(device);

	start_addr = sector * MMCSD_FS_SECTOR_SIZE;

	g_selected_lock[device->port].device = device;
	TRACE("read from 0x%0lX, number:0x%0lX\n",
		(unsigned long)sector, (unsigned long)nr_sector);

	state = MMCSD_TRANSFER;
	ret = mmcsd_check_status(device,&state);
	if(ret)
	{
		FAILED("Check Status failed!\n");
		goto read_fail;
	}

	/********************************************************************
	 * Get the Lock for seclected device
	 *******************************************************************/

	ret = _MMCSD_SetBlockLen(device,MMCSD_MINIMUM_BLK_SIZE,1);
	if(ret)
	{
		FAILED("Set Block Len failed, ret = 0x%x\n",ret);
		goto read_fail;
	}

	state = MMCSD_TRANSFER;
	ret = mmcsd_check_status(device,&state);


	/********************************************************************
	 * Begin to Read
	 *******************************************************************/

	if((ret=(mmcsd_ReadBlock(device,start_addr, nr_sector, buffer)))!=MMCSD_OK)
	{
		FAILED(" ReadBlock error\n");
		device->card_state = MMCSD_NOTREADY;
		goto read_fail;
	}

	device->card_state = MMCSD_STANDBY;

	mmcsd_device_sem_release(device);

	return MMCSD_OK;

read_fail:
	mmcsd_device_sem_release(device);

	return MMCSD_ERROR;
}

/******************************************************************************
 * Function Name: mmcsd_write_device
 *
 * Input:
 * 		device: the device write to
 * 		buffer:	memory area for I/O transfer.
 * 		sector: first sector number;
 * 		nr_sector: number of the sectos
 *
 * Value Returned:
 * 		0:	success.
 * 		1:	failure.
 *
 * Description:
 * 	this routine will write data into indicated sectors.
 *
 * Modification History:
 * 	30 DEC,2001, Initial version.  			Yiwei Zhao
 *****************************************************************************/

int mmcsd_write_device(	MMCSD_Dev *device, char * buffer,
		uint32_t sector, uint32_t nr_sector)
{
	int ret;

	TRACE(" write to 0x%lx, number:0x%lx\n", (unsigned long)sector, (unsigned long)nr_sector);

	device->addr = sector * device->blksize;

	while (nr_sector)
	{
		int nr_write = MMCSD_MAX_MULTI_BLK - device->cur_num;
		if(nr_write > nr_sector)
			nr_write = nr_sector;

#ifdef MMCSD_DMA_ENDIAN_ERR
		_mmcsd_cp_16_little_2_big(buffer, device->tmp_pool[device->cur_num], nr_write * device->blksize);
#else
		memcpy(device->buff_pool[device->cur_num], buffer, nr_write * device->blksize);
#endif
		if (device->cur_num == 0)

		device->cur_num += nr_write;
		device->cur_sec += nr_write;

		buffer += nr_write * device->blksize;
		nr_sector -= nr_write;

		if ((nr_sector == 0) || (device->cur_num == MMCSD_MAX_MULTI_BLK))
		{
#ifdef MMCSD_WRITE_DMA
			ret = mmcsd_dma_write_blks(device);
#else
			ret = mmcsd_write_blks(device);
#endif
			if (ret != MMCSD_OK)
			{
				FAILED("mmcsd_write_blk failed\n");
				goto write_fail;
			}
			// cur_num should be zero
		}
	}

	return MMCSD_OK;

write_fail:
	mmcsd_device_sem_release(device);

	return ret;
}

MMCSD_STATUS mmcsd_write_blks(MMCSD_Dev *device)
{

	int i,t,j,counter;
	uint32_t state,ret,ret2,int_mask, temp32;
	uint16_t temp16,*p16;
	uint32_t addr;
	uint32_t	burst_len=8;

	if(device->cur_num == 0)
	{
		mmcsd_free_buffer(device);
		return MMCSD_OK;
	}

	ret=ret2=MMCSD_OK;

	TRACE("enter mmcsd_write_blks.....\n");

	g_selected_lock[device->port].device = device;

	/**************************************************************
	 * Check the Card Status
	 **************************************************************/
	state =MMCSD_TRANSFER;
	ret = mmcsd_check_status(device,&state);
	if(ret)
	{
		mmcsd_free_buffer(device);
		return ret;
	}

	/**************************************************************
	 * Mask Interrupt
	 **************************************************************/
	TRACE("Disable Interrupt..\n");
	_MMCSD_Mask_interrupt(device,0x7F,&int_mask);
	/**
	 * End of Testing
	 */

	/**************************************************************
	 * Set Block Length
	 **************************************************************/
	i = 5;
	while(i--)
	{
		ret = _MMCSD_SetBlockLen(device,MMCSD_MINIMUM_BLK_SIZE,device->cur_num);
		if(ret)
		{
			TRACE("3)Set Block Len, ret = 0x%lx,state=%lx\n",(unsigned long)ret,(unsigned long)state);
			TRACE("4)Try again...%d\n",i);
			state = MMCSD_TRANSFER;
			mmcsd_check_status(device,&state);
		}
		else	//ret == SUCCESS!
		{
			break;
		}

	}
	if(ret)
	{
		mmcsd_free_buffer(device);
		FAILED("5)Set Blocked Len Failed!\n");
		return ret;

	}
	/********************************************************************
	 * Begin to write
	 *******************************************************************/


	addr = device->addr;
	TRACE("6)addr 0x%lx (sec 0x%lx), blks %ld\n",
			(unsigned long)addr,(unsigned long)(addr/512),(unsigned long)(device->cur_num));

	_MMCSD_StopClk(device);

	device->pending_irq = 0;

	if(device->cur_num == 1)
	{
		if(device->bit_width == 0)
		{
			_MMCSD_SetCmd(device,MMCSD_CMD24,addr,MMCSDB_DATEN|MMCSDB_R1|MMCSDB_WRRD);
			burst_len = 8;
		}
		else
		{
			_MMCSD_SetCmd(device,MMCSD_CMD24,addr,MMCSD_BUS_4BIT|MMCSDB_DATEN|MMCSDB_R1|MMCSDB_WRRD);
			burst_len = 32;
		}
		device->cur_cmd = MMCSD_CMD24;
	}
	else
	{
		if(device->bit_width == 0)
		{
			_MMCSD_SetCmd(device,MMCSD_CMD25,addr,MMCSDB_DATEN|MMCSDB_R1|MMCSDB_WRRD);
			burst_len = 8;
		}
		else
		{
			_MMCSD_SetCmd(device,MMCSD_CMD25,addr,MMCSD_BUS_4BIT|MMCSDB_DATEN|MMCSDB_R1|MMCSDB_WRRD);
			burst_len = 32;
		}
		device->cur_cmd = MMCSD_CMD25;
	}
	_MMCSD_StartClk(device);

	ret = MMCSD_OK;
	for( t=0; t<device->cur_num; t++)
	{
		counter = 0;
#ifdef MMCSD_DMA_ENDIAN_ERR
		p16 = (uint16_t *)device->tmp_pool[t];
#else
		p16 = (uint16_t *)device->buff_pool[t];
#endif
		while(ret == MMCSD_OK)
		{
			uint32_t status;

			status = mmcsd_reg_rd(device, MX1_MMCSD_STATUS_o);

			if((status&0x1043)==0)
			{
				continue;
			}

			j = 0xFF;
			while(j--);
			//			TRACE("MMCSD_STATUS = 0x%x\n",status);
			//
			if(status&0x0003)
			{
				TRACE("7)ERROR!--, MMCSD_STATUS = %lx\n",(unsigned long)status);
				ret = MMCSD_ERROR;
			}
			else if((status&0x0040)!=0) // Buffer is empty
			{
				//				TRACE("FIFO is empty!\n");
				for(i = 0; i < burst_len; i++)
				{
					temp16 = (*p16);
					temp32 = temp16;

					mmcsd_reg_wr(device, MX1_MMCSD_BUFFER_ACCESS_o, temp32);

					p16++;

				}
				counter += burst_len*2;
			}
			if(counter == device->blksize)
			{
				break;
			}
		}
		if(ret!=MMCSD_OK)
			break;
	}

	/*******************************************************
	 * Write Done
	 *******************************************************/
	if(ret==MMCSD_OK)
	{
		while(1)
		{
			uint32_t status;

			status = mmcsd_reg_rd(device, MX1_MMCSD_STATUS_o);

			if(status&0x0007)
			{
				TRACE("8)ERROR! status = 0x%lx\n",(unsigned long)status);
				ret = MMCSD_ERROR;
				break;
			}
			else if((status&0x1000)!=0)
			{
				TRACE("9)Write Done! MMCSD_STATUS =0x%lx\n",(unsigned long)status);
				ret = MMCSD_OK;
				break;
			}
			TRACE("10)MMCSD_STATUS =%lx\n",(unsigned long)status);
		}
	}

	TRACE("Enable Intr...\n");
	_MMCSD_Restore_Mask(device,int_mask);
	device->cur_cmd = 0;

	if(device->cur_num>1)
	{
		ret2=_MMCSD_NoDataCommand(device,MMCSD_CMD12,device->card_rca<<16,MMCSDB_R1);
	}

	//	state =MMCSD_TRANSFER;
	//	mmcsd_check_status(device,&state);

	mmcsd_free_buffer(device);
	TRACE("15)return with ret = %lx,ret2 = %lx\n",(unsigned long)ret,(unsigned long)ret2);
	return(ret|ret2);
}

/*!
 * Dump binary data in hex
 */
static void kdump(const void *_buf, size_t len)
{
	int i;
	const unsigned char *buf = (const unsigned char *)_buf;

	for (i = 0; i < len; i++)
	{
		printf("%02x ", *buf++);
		if (i % 16 == 15)
			printf("\n");
	}
	printf("\n");
}

#ifdef MMCSD_WRITE_DMA
MMCSD_STATUS mmcsd_dma_write_blks(MMCSD_Dev *device)
{
	int i,timeleft;
	uint32_t state,ret,ret2,int_mask = 0xFFFFFFFF;
	uint32_t addr;
	uint32_t	status,counter;


	if(device->cur_num == 0)
	{
		mmcsd_free_buffer(device);
		return MMCSD_OK;
	}


	ret=ret2=MMCSD_OK;
	g_selected_lock[device->port].device = device;

	/**************************************************************
	 * Check the Card Status
	 **************************************************************/
	//TRACE("Check Status\n");

	state =MMCSD_TRANSFER;
	ret = mmcsd_check_status(device,&state);
	if(ret)
	{
		mmcsd_free_buffer(device);
		FAILED("2)Check Error!\n");
		return ret;
	}

	/**
	 * End of Testing******************************************************************
	 */
	/**************************************************************
	 * Set Block Length
	 **************************************************************/
	i = 5;
	while(i--)
	{
		//	TRACE("Set Block Length\n");
		ret = _MMCSD_SetBlockLen(device,MMCSD_MINIMUM_BLK_SIZE,device->cur_num);
		if(ret)
		{
			TRACE("3)Set Block Len, ret = 0x%lx,state=%lx\n",(unsigned long)ret,(unsigned long)state);
			TRACE("4)Try again...%d\n",i);
			state = MMCSD_TRANSFER;
			mmcsd_check_status(device,&state);
		}
		else	//ret == SUCCESS!
		{
			break;
		}
	}
	if(ret)
	{
		mmcsd_free_buffer(device);
		FAILED("5)Set Blocked Len Failed!\n");
		goto dma_write_end;
	}

	state = MMCSD_TRANSFER;
	ret = mmcsd_check_status(device,&state);
	if(ret)
	{
		mmcsd_free_buffer(device);
		FAILED("Check Status failed.\n");
		goto dma_write_end;
	}

	i = 10;
	while(i--)
	{
	}
	if(ret)
	{
		mmcsd_free_buffer(device);
		FAILED("Set Bus Width Failed!\n");
		goto dma_write_end;
	}
	/********************************************************************
	 * Begin to write
	 *******************************************************************/

	addr = device->addr;
	TRACE("*** addr 0x%lx (sect 0x%lx, %ld), %ld blks\n",
			(long)addr,(long)addr/512,(long)addr/512,(long)device->cur_num);

	//kdump(device->buff_pool[0], device->cur_num * MMCSD_MINIMUM_BLK_SIZE);

	/**************************************************************
	 * Mask Interrupt
	 **************************************************************/
	//TRACE("Disable Interrupt\n");
	_MMCSD_Mask_interrupt(device,0x78,&int_mask);
	/*
	 *  DMA Preparation
	 */
	/**************************************************************
	 * Config DMA
	 **************************************************************/

	//	TRACE("DMA Configuration\n");
	ret = mmcsd_DmaStartTransfer(device,
#ifdef MMCSD_DMA_ENDIAN_ERR
			device->tmp_pool[0],
#else
			device->buff_pool[0],
#endif
			MMCSD_MINIMUM_BLK_SIZE*(device->cur_num),
			MMCSD_WRITE,
			device->bit_width);
	if(ret != MMCSD_OK)
	{
		FAILED("mmcsd_DmaStartTransfer failed\n");
		goto dma_write_end;
	}


	/**************************************************************
	 * Set Command
	 **************************************************************/

	state =MMCSD_TRANSFER;
	ret = mmcsd_check_status(device,&state);
	if(ret)
	{
		FAILED("Check Status failed.\n");
	}


	//	TRACE("Set Command (24/25)\n");

	_MMCSD_StopClk(device);

	device->pending_irq = 0;

	if(device->cur_num == 1)
	{
		if(device->bit_width == 0)
		{
			_MMCSD_SetCmd(device,MMCSD_CMD24,addr,MMCSDB_DATEN|MMCSDB_R1|MMCSDB_WRRD);
		}
		else
		{
			_MMCSD_SetCmd(device, MMCSD_CMD24,addr,MMCSD_BUS_4BIT|MMCSDB_DATEN|MMCSDB_R1|MMCSDB_WRRD);
		}
	}
	else
	{
		if(device->bit_width == 0)
		{
			_MMCSD_SetCmd(device, MMCSD_CMD25,addr,MMCSDB_DATEN|MMCSDB_R1|MMCSDB_WRRD);
		}
		else
		{
			_MMCSD_SetCmd(device, MMCSD_CMD25,addr,MMCSD_BUS_4BIT|MMCSDB_DATEN|MMCSDB_R1|MMCSDB_WRRD);
		}
	}

	//	TRACE("Start Clk... Wait CMD Done.\n");
	ret = MMCSD_OK;
	_MMCSD_StartClk(device);

	//if((ret=_MMCSD_irq_WaitUntil(MMCSD_CMD_DONE,MMCSD_CMD25,1000))!=0)
	if((ret=_MMCSD_WaitUntil(device, MMCSD_CMD_DONE))!=MMCSD_OK)
	{
		FAILED("WaitCMDDone ret 0x%lx\n",(unsigned long)ret);
		goto dma_write_end;
	}

	status = mmcsd_reg_rd(device, MX1_MMCSD_STATUS_o);

	if((status&0x40)==0)
	{
		FAILED("FIFO is not empty  0x%lx\n",(unsigned long)ret);
		goto dma_write_end;
	}

	/**************************************************************
	 * Enable DMA
	 **************************************************************/

	//	TRACE("Enable DMA..\n");
	//	_MMCSD_StartClk();
	device->cur_cmd = 0xFF;
#ifdef CONFIG_ARCH_MX1ADS
	enable_dma(device->dma_channel);
#endif
#ifdef CONFIG_ARCH_MX2ADS
	mx_dma_start(device->dma_channel);
#endif

	timeleft = mmcsd_device_select_wait(device,1000);
	if(!timeleft)
	{
		FAILED("write failed, Timeout!\n");
#ifdef CONFIG_ARCH_MX1ADS
		disable_dma(device->dma_channel);
#endif
#ifdef CONFIG_ARCH_MX2ADS
		_reg_DMA_CCR(device->dma_channel) &= 0xFFFFFFFE;
#endif
		device->cur_cmd = 0x0;
		ret = MMCSD_ERROR;
		goto dma_write_end;
	}
#ifdef CONFIG_ARCH_MX1ADS
	disable_dma(device->dma_channel);
#endif
#ifdef CONFIG_ARCH_MX2ADS
	_reg_DMA_CCR(device->dma_channel) &= 0xFFFFFFFE;
#endif
	device->cur_cmd = 0x0;
	//	TRACE("Waked up\n");

	status = device->result;
	TRACE("waked up! status = %lx\n",(long)status);

	counter = 0xFFFF;
	while(counter--)
	{
		status =mmcsd_reg_rd(device, MX1_MMCSD_STATUS_o);

		TRACE("MMCSD_STATUS is 0x%0lX (for TRANS-DONE)\n",(unsigned long)status);
		if((status&0x1007)==0)
		{
			mmcsd_delay(0xFFFF);
			continue;
		}

		if(status&0x0007)
		{
			ret = MMCSD_ERROR;
			FAILED("status is 0x%lx\n",(unsigned long)status);
			goto dma_write_end;
		}
		ret = MMCSD_OK;
		break;
	}
	if(counter == 0)
	{
		ret = MMCSD_ERROR;
		FAILED("TIME OUT..\n");
		goto dma_write_end;
	}

	/*******************************************************
	 * Write Done
	 *******************************************************/
	if(ret==MMCSD_OK)
	{
		counter = 0xFFFF;
		while(counter--)
		{
			status = mmcsd_reg_rd(device, MX1_MMCSD_STATUS_o);

			TRACE("MMCSD_STATUS is 0x%0lX (for Write-DONE)\n",(unsigned long)status);
			if(status&0x0004)
			{
				TRACE("8)ERROR!\n");
				ret = MMCSD_ERROR;
				break;
			}
			else if((status&0x1000)!=0)
			{
				ret = MMCSD_OK;
				break;
			}
			mmcsd_delay(0xFFFF);
		}
		if(counter==0)
		{
			ret = MMCSD_ERROR;
			FAILED("Time Out for Write-DONE.\n");
			goto dma_write_end;

		}
	}

dma_write_end:
	device->cur_cmd = 0;

	if((ret == MMCSD_OK)&&(device->cur_num>1))
	{
		MMCSD_Responses rp;
		//		TRACE("CMD12 -- Last Block.\n");
		ret2=_MMCSD_NoDataCommand(device, MMCSD_CMD12,0,MMCSDB_R1);
		_MMCSD_GetResponses(device, &rp,MMCSDB_R1);
	}

	//	INFO("Check Status.....(after cmd12)");
	state =MMCSD_TRANSFER;
	mmcsd_check_status(device,&state);

	if(int_mask != 0xFFFFFFFF)
	{
		//		TRACE("Enable Interrupt.\n");
		_MMCSD_Restore_Mask(device, int_mask);
	}

	mmcsd_free_buffer(device);
	//	TRACE("15)return with ret = %lx,ret2 = %lx\n",ret,ret2);
	return(ret|ret2);
}
#endif /*MMCSD_WRITE_DMA*/


void mmcsd_free_buffer(MMCSD_Dev *device)
{

	device->cur_num = 0;
	device->addr = 0;
}

/******************************************************************************
 * Function Name: mmcsd_sector_erase
 *
 * Input:
 * 		device: the related device
 * 		blks:	the blocks that will be erased.
 *
 * Value Returned:
 * 		0:	success.
 * 		1:	failure.
 *
 * Description:
 * 	this routine will erase indicated group.
 *
 * Modification History:
 * 	30 DEC,2001, Initial version.  			Yiwei Zhao
 *****************************************************************************/

static int mmcsd_sector_erase(MMCSD_Dev *device, MMCSD_Blks *blks)
{
	int ret;
	uint32_t state;

	TRACE("Erase the sector  %ld\n", (long int)(blks->start));

	/********************************************************************
	 * Lock the device.
	 *******************************************************************/
#ifndef MMCSD_TEST
	mmcsd_device_sem_obtain(device);
	g_selected_lock[device->port].device = device;
#endif

	/********************************************************************
	 * Identify the Card State ( Process only for Stand By State)
	 *******************************************************************/
	state =MMCSD_TRANSFER;
	ret = mmcsd_check_status(device,&state);
	if(ret)
	{
		FAILED(" Check Status failed.\n");
		goto sector_erase_ret;
	}

	/********************************************************************
	 * Erase
	 *******************************************************************/

	ret = _MMCSD_TagSectorStart(device,blks->start);
	if(ret)
	{
		TRACE("TagSectorStart error\n");
		goto sector_erase_ret;
	}

	//ret = _MMCSD_TagSectorEnd(device,blks->start);
	ret = _MMCSD_TagSectorEnd(device,blks->start+device->phy_sectorsize-1);
	if(ret)
	{
		TRACE("TagSectorEnd error\n");
		goto sector_erase_ret;
	}

	ret = _MMCSD_Erase(device);
	if(ret)
	{
		TRACE("Erase error\n");
		goto sector_erase_ret;
	}

sector_erase_ret:
	/********************************************************************
	 * Unlock the selected device structure.
	 *******************************************************************/
	state =MMCSD_TRANSFER;
	mmcsd_check_status(device,&state);
#ifndef MMCSD_TEST
	mmcsd_device_sem_release(device);
#endif


	return ret;
}

/******************************************************************************
 * Function Name: mmcsd_ReadBlock
 *
 * Input:               None.
 *
 * Value Returned:
 *
 * 		MMCSD_STATUS
 *
 * Description:
 *
 *
 * Modification History:
 *      30 DEC,2001, Initial version.                   Yiwei Zhao
 *****************************************************************************/
MMCSD_STATUS mmcsd_ReadBlock(MMCSD_Dev *device,uint32_t start_addr,uint16_t nsects, uint8_t *buff )
{
	uint16_t  i;
	uint32_t  read_addr;
	MMCSD_STATUS ret;
	uint8_t	*cur_buf;
	/************************************************************************
	 *  Initialize local variables
	 *************************************************************************/
	ret = read_addr = 0;                            /* SEND/RECEIVE ERROR   */

	for ( i = 0 ; i < nsects ; i++ )
	{
		/************************************************************
		 *  Set the first address of read sector
		 **************************************************************/
		read_addr = ( start_addr + (MMCSD_MINIMUM_BLK_SIZE*i) );
		/************************************************************
		 *  Create cmd
		 **************************************************************/
		cur_buf = buff+(MMCSD_MINIMUM_BLK_SIZE*i);

#ifdef MMCSD_READ_DMA

		ret=mmcsd_DMA_ReadSingleBlock(device,read_addr, cur_buf);
		if(ret==MMCSD_OK)	// Only for DMA
		{			// Endian processing
#ifdef MMCSD_DMA_ENDIAN_ERR
			_mmcsd_16_big_2_little(cur_buf,MMCSD_MINIMUM_BLK_SIZE);
#endif
		}
#else
		ret=mmcsd_ReadSingleBlock(device,read_addr, cur_buf);

#endif
		if (ret)
		{
			break;
		}
	}
	return( ret );
}

#ifdef MMCSD_READ_DMA
/******************************************************************************
 * Function Name: mmcsd_ReadSingleBlock
 *
 * Input:               None.
 *
 * Value Returned:
 *
 * 	MMCSD_STATUS
 *
 * Description:
 *
 * Modification History:
 *      30 DEC,2001, Initial version.                   Yiwei Zhao
 *****************************************************************************/
MMCSD_STATUS mmcsd_DMA_ReadSingleBlock(MMCSD_Dev *device,uint32_t addr,uint8_t * buff)
{
	MMCSD_STATUS ret;
	uint32_t channel,status,int_mask,state;
	int timeleft;

	device->result = MMCSD_OK;
	channel = device->dma_channel;

	ret = mmcsd_DmaStartTransfer( device,
			buff,
			MMCSD_MINIMUM_BLK_SIZE,
			MMCSD_READ,
			device->bit_width);
	if(ret != MMCSD_OK)
	{
		FAILED("mmcsd_DmaStartTransfer failed !\n");
		return ret;
	}

	state = MMCSD_TRANSFER;
	ret = mmcsd_check_status(device,&state);

	_MMCSD_Mask_interrupt(device, 0x7F,&int_mask);

	_MMCSD_StopClk(device);

	device->pending_irq = 0;

	if(device->bit_width)
	{
		_MMCSD_SetCmd(device, MMCSD_CMD17,addr,MMCSD_BUS_4BIT|MMCSDB_DATEN|MMCSDB_R1);	// Read Single BLK
	}
	else
	{
		_MMCSD_SetCmd(device, MMCSD_CMD17,addr,MMCSDB_DATEN|MMCSDB_R1);	// Read Single BLK
	}
	device->cur_cmd = MMCSD_CMD17;

	_MMCSD_StartClk(device);
	if((ret=_MMCSD_WaitUntil(device, MMCSD_CMD_DONE))!=0)
	{
		FAILED("WaitCMDDone ret 0x%lx\n",(unsigned long)ret);
		goto dma_read_end;
	}
	TRACE("Enable DMA..\n");
#ifdef CONFIG_ARCH_MX1ADS
	enable_dma(channel);
#endif
#ifdef CONFIG_ARCH_MX2ADS
	mx_dma_start(channel);
#endif

	timeleft = mmcsd_device_select_wait(device,1000);
	if(!timeleft)
	{
		TRACE("read failed, Timeout!\n");
		ret = MMCSD_ERROR;
		goto dma_read_end;
	}

	status = device->result;

	TRACE("waked up! status = %x\n", status);
	if(status&0x800)
	{
		ret = MMCSD_OK;
		goto dma_read_end;
	}
	else if(status&0x3)
	{
		TRACE("MMCSD_STATUS 0x%lx\n",(unsigned long)status);
		ret = MMCSD_ERROR;
		goto dma_read_end;
	}
	// in fact, although status&0x800 ==0, read succeed.

dma_read_end:
#ifdef CONFIG_ARCH_MX1ADS
	disable_dma(channel);
#endif
#ifdef CONFIG_ARCH_MX2ADS
	//disable dma
	_reg_DMA_CCR(channel) &= 0xFFFFFFFE;
#endif
	_MMCSD_Restore_Mask(device, int_mask);

	device->cur_cmd = 0;
	return ret;

}
#endif /*MMCSD_READ_DMA*/

MMCSD_STATUS mmcsd_ReadSingleBlock(MMCSD_Dev *device,uint32_t addr,uint8_t * buff)
{
	MMCSD_STATUS ret;
	uint32_t temp32,j,int_mask,total;
	uint16_t temp16;
	uint8_t *pbuff;
	int i;

	pbuff=buff;

	_MMCSD_Mask_interrupt(device, 0x7F,&int_mask);

	total = 0;
	_MMCSD_StopClk(device); 	//Ricky

	device->pending_irq = 0;

	if(device->bit_width)
	{
		_MMCSD_SetCmd(device, MMCSD_CMD17,addr,MMCSD_BUS_4BIT|MMCSDB_DATEN|MMCSDB_R1);	// Read Single BLK
	}
	else
	{
		_MMCSD_SetCmd(device, MMCSD_CMD17,addr,MMCSDB_DATEN|MMCSDB_R1);	// Read Single BLK
	}
	device->cur_cmd = MMCSD_CMD17;
	_MMCSD_StartClk(device);

	while(1)
	{
		uint32_t status;

		status =mmcsd_reg_rd(device, MX1_MMCSD_STATUS_o);

		if((status&0x088B) == 0)
		{
			continue;
		}

		j = 0xFF;
		while(j--);

		if(status&0x000B)
		{
			FAILED("Read Error (MMCSD_STATUS = 0x%lx\n",(unsigned long)status);
			ret = MMCSD_ERROR;
			goto signle_read_end;
		}
		/*		if(status&0x0008)
				{
				goto single_read_try_again;
				}
		 */
		if((status&0x0080)!=0)		// Buffer is FULL
		{
			//TRACE("FIFO is full\n");
			if(device->bit_width == 0)
			{
				for(i = 0;i < 8; i++)
				{
					// Aleph: certainly a Motorola pasto
					// status =mmcsd_reg_rd(device, MX1_MMCSD_STATUS_o);
					temp32 = mmcsd_reg_rd(device, MX1_MMCSD_BUFFER_ACCESS_o);

					temp16 = (uint16_t)temp32;
#ifdef MMCSD_DMA_ENDIAN_ERR
					*pbuff = (temp16&0xFF00)>>8;
					pbuff++;
					*pbuff = (temp16&0x00FF);
					pbuff++;
#else
					*pbuff = (temp16&0x00FF);
					pbuff++;
					*pbuff = (temp16&0xFF00)>>8;
					pbuff++;
#endif
				}
				total += 16;
			}
			else
			{
				for(i = 0;i < 32; i++)
				{
					temp32 = mmcsd_reg_rd(device, MX1_MMCSD_BUFFER_ACCESS_o);

					temp16 = (uint16_t)temp32;
#ifdef MMCSD_DMA_ENDIAN_ERR
					*pbuff = (temp16&0xFF00)>>8;
					pbuff++;
					*pbuff = (temp16&0x00FF);
					pbuff++;
#else
					*pbuff = (temp16&0x00FF);
					pbuff++;
					*pbuff = (temp16&0xFF00)>>8;
					pbuff++;
#endif
				}
				total += 64;
			}
		}

		if((status&0x0800) != 0)	// TRANS Done
		{
			//TRACE("Transfer done! MMCSD_STATUS = 0x%x\n",status);
			ret = MMCSD_OK;
			goto signle_read_end;

		}

		//		TRACE("Sleep on.. status=%x\n",status);
		//		timeleft = mmcsd_device_select_wait(device,100000);
		//		TRACE("Waked up ! status = 0x%x\n",status);
		//
	}


signle_read_end:

	_MMCSD_StopClk(device);
	_MMCSD_Restore_Mask(device, int_mask);

	device->cur_cmd = 0;
	return ret;

}

/******************************************************************************
 * Function Name: mmcsd_DmaStartTransfer
 *
 * Input:
 *
 * Value Returned:
 *
 * Description:
 *
 *
 * Modification History:
 *      30 DEC,2001, Initial version.                   Yiwei Zhao
 *****************************************************************************/

int mmcsd_DmaStartTransfer(MMCSD_Dev *device,uint8_t * buff,uint32_t counter, uint8_t mode,uint8_t is_4_bit)
{
	int channel;

	channel = device->dma_channel;
	if(mode == MMCSD_READ)
	{
#ifdef CONFIG_ARCH_MX1ADS
		_reg_DMA_DAR(channel) = (uint32_t)__virt_to_phys(buff);
		/*FIXME: is the masking of MSB four bits intentional*/
		_reg_DMA_SAR(channel) = (device->base_addr+MX1_MMCSD_BUFFER_ACCESS_o)&0x0FFFFFFF;
		_reg_DMA_CNTR(channel) = counter;

		_reg_DMA_CCR(channel) = 0x00000828;	// Source Size = 16,Des Len = 32
		_reg_DMA_RSSR(channel) = 0xD;
		if(is_4_bit)
		{
			_reg_DMA_BLR(channel) = 0x0;			// Burst Len 64;
			//_reg_DMA_BLR(channel) = 0x20;			// Burst Len 32;
		}
		else
		{
			_reg_DMA_BLR(channel) = 0x10;			// Burst Len 16;
		}
#endif
#ifdef CONFIG_ARCH_MX2ADS
		(device->dma).sourceType = DMA_TYPE_FIFO;
		(device->dma).sourcePort = DMA_MEM_SIZE_16;
		(device->dma).destType = DMA_TYPE_LINEAR;
		(device->dma).destAddr = (uint32_t)__virt_to_phys(buff);
		(device->dma).destPort = DMA_MEM_SIZE_32;
		(device->dma).count = counter;
		(device->dma).sourceAddr = device->base_addr+MX1_MMCSD_BUFFER_ACCESS_o;
		if(device->port == 0)
		{
			/*(device->dma).sourceAddr = 0x10013038;//sdhc1,*/
			(device->dma).request = 0x7;//sdhc1
		}
		else
		{
			/*(device->dma).sourceAddr = 0x10014038;// for sdhc2 is 0x10014038*/
			(device->dma).request = 0x6;// sdhc2 is 0x6;
		}

		if(is_4_bit)
		{
			(device->dma).burstLength = 0x0; //burst length 64
		}
		else
		{
			(device->dma).burstLength = 0x10; //burst len 16
		}
		mx_dma_set_config(device->dma_channel,&(device->dma));
		//		mx_dump_dma_register(device->dma_channel);

#endif
	}
	else
	{
#ifdef CONFIG_ARCH_MX1ADS
		//	_reg_DMA_SAR(channel) = (uint32_t)_MMCSD_get_phy_addr(buff);
		_reg_DMA_SAR(channel) = (uint32_t)__virt_to_phys(buff);
		/*FIXME: is the masking of MSB four bits intentional*/
		_reg_DMA_DAR(channel) = (device->base_addr+MX1_MMCSD_BUFFER_ACCESS_o)&0x0FFFFFFF;
		_reg_DMA_CNTR(channel) = counter;

		_reg_DMA_CCR(channel) = 0x2088;	// Source Size = 32,Des Len = 16
		_reg_DMA_RSSR(channel) = 0xD;

		if(is_4_bit)
		{
			_reg_DMA_BLR(channel) = 0x0;		// Burst Len 64;
			//_reg_DMA_BLR(channel) = 0x20;		// Burst Len 32;
		}
		else
		{
			_reg_DMA_BLR(channel) = 0x10;		// Burst Len 16;
			//	_reg_DMA_BLR(channel) = 0x8;		// Burst Len 8;
		}
#endif
#ifdef CONFIG_ARCH_MX2ADS
		(device->dma).sourceType = DMA_TYPE_LINEAR;
		(device->dma).sourceAddr = (uint32_t)__virt_to_phys(buff);
		(device->dma).sourcePort = DMA_MEM_SIZE_32;

		(device->dma).destType = DMA_TYPE_FIFO;
		(device->dma).destPort = DMA_MEM_SIZE_16;

		(device->dma).count = counter;
		(device->dma).sourceAddr = device->base_addr+MX1_MMCSD_BUFFER_ACCESS_o;
		if(device->port == 0)
		{
			/*(device->dma).destAddr = 0x10013038;//sdhc1*/
			(device->dma).request = 7;//sdhc1
		}
		else
		{
			/*(device->dma).destAddr = 0x10014038;//for sdhc2 is 0x10014038*/
			(device->dma).request = 0x6;//sdhc2 is 0x6;
		}

		if(is_4_bit)
		{
			(device->dma).burstLength = 0x0; //burst length 64
		}
		else
		{
			(device->dma).burstLength = 0x10; //burst len 16
		}

		mx_dma_set_config(device->dma_channel,&(device->dma));

#endif
	}

	return MMCSD_OK;
}

/******************************************************************************
 * Function Name: mmcsd_dma_handler
 *
 * Input:
 *
 * Value Returned:
 *
 * Description:
 *
 *
 * Modification History:
 *      30 DEC,2001, Initial version.                   Yiwei Zhao
 *****************************************************************************/

void mmcsd_dma_isr(void *arg)
{
	MMCSD_Dev *device;
	uint32_t status;

	device = g_selected_lock[0].device;

	if(device == NULL)
	{
		TRACE("mmcsd_interrupt_handler: no device\n");
		return;
	}

	status =mmcsd_reg_rd(device, MX1_MMCSD_STATUS_o);


	//TRACE("(status = 0x%x)\n",status);

	//	if(status&0x803)   // It should be here, but sometimes no 0x800 return.
	{
		device->result = status;
		mmcsd_device_select_wake(device);
		//	TRACE("Wake Up ...\n");
	}

}

#ifdef CONFIG_ARCH_MX2ADS
void mmcsd_dma_isr_1(void * arg)
{
	MMCSD_Dev *device;
	uint32_t status;

	device = g_selected_lock[1].device;

	if(device == NULL)
	{
		TRACE("\nmmcsd_interrupt_handler: no device\n");
		return;
	}

	status =mmcsd_reg_rd(device, MX1_MMCSD_STATUS_o);


	//TRACE("(status = 0x%x)\n",status);

	//	if(status&0x803)   // It should be here, but sometimes no 0x800 return.
	{
		device->result = status;
		mmcsd_device_select_wake(device);
		//	TRACE("Wake Up ...\n");
	}

}
#endif

void mmcsd_dma_err_isr(int error_type)
{
	MMCSD_Dev *device;
	uint32_t status;

	device = g_selected_lock[0].device;

	if(device == NULL)
	{
		TRACE("mmcsd_interrupt_handler: no device\n");
		return;
	}

	status =mmcsd_reg_rd(device, MX1_MMCSD_STATUS_o);

	device->result = status;
	TRACE("error_type = 0x%x\n",error_type);

	mmcsd_device_select_wake(device);

}

#ifdef MMCSD_WITH_RTEMS
rtems_isr mmcsd_irq_handler(rtems_irq_hdl_param data)
#else /*MMCSD_WITH_RTEMS*/
void mmcsd_irq_handler(int irq,void * data, struct pt_regs * pt)
#endif /*MMCSD_WITH_RTEMS*/
{
	MMCSD_Dev *device=(MMCSD_Dev *)data;
	uint32_t status;

	if(device == NULL)
	{
		return;
	}

	g_mmcsd_irq_count++;

	mmcsd_reg_set(device, MX1_MMCSD_INT_MASK_o, 0x40);

	//	TRACE("The Device not empty...\n");
	status = mmcsd_reg_rd(device, MX1_MMCSD_STATUS_o);
	//TRACE("irq is called (%d), MMCSD_STATUS = 0x%0X\n",g_mmcsd_cmd,status);

	device->result = status;
	device->pending_irq = 1;

	//	TRACE("Wake up the deice.. (status 0x%x)\n",device->result);

	if(device->cur_cmd != 0xFF)
	{
		mmcsd_device_select_wake(device);
	}

}
/******************************************************************************
 * Function Name: mmcsd_tasklet
 *
 * Input:
 *
 * Value Returned:
 *
 * Description:
 *
 *
 * Modification History:
 *      30 DEC,2001, Initial version.                   Yiwei Zhao
 *****************************************************************************/
/*
   void mmcsd_tasklet()
   {
   MMCSD_Dev * device;

   device = g_selected_lock.device;

   if(device == NULL)
   {
   TRACE("\n mmcsd_tasklet: no device\n");
   return;
   }

   switch (g_mmcsd_cmd)
   {

   case MMCSD_CMD17:

   case MMCSD_CMD24:

   disable_dma(MMCSD_DMA_CHANNEL);

   device->result = get_dma_residue(MMCSD_DMA_CHANNEL);

   free_dma(MMCSD_DMA_CHANNEL);

   mmcsd_device_select_wake(device);

   break;
   default:
   break;

   }
   }
 */
/******************************************************************************
 * Function Name: mmcsd_ioctl_device
 *
 * Input:
 * 		inode	:	the pointer to MS driver-related inode.
 *		filep	:	the pointer to MS driver-related file structure.
 *		command	:	the command number.
 *		argument:	argument which depends on command.
 *
 * Value Returned:
 *		0	:	succeed
 *		1	:	fail
 *
 * Description:
 * 	This routine will implement driver-specific functions.
 *
 * Modification History:
 * 	30 DEC,2001,  Yiwei  Zhao
 *****************************************************************************/

int mmcsd_ioctl_device(MMCSD_Dev *device,
				unsigned command,
				unsigned long  argument)
{
	int ret;
	MMCSD_PWD *pwd;
	MMCSD_Blks blks;
	uint32_t 	addr;

	TRACE("mmcsd_ioctl is called!\n");

	switch(command) {


#if 0
		case BLKFLSBUF:
		case BLKRAGET:
			return blk_ioctl(inode->i_rdev, command, argument);

		case BLKGETSIZE:

			if (!argument) return -EINVAL; /* NULL pointer: not valid */

			if (copy_to_user((long *) argument, &(device->dev_size), sizeof (long)))
			{
				return -EFAULT;
			}
			return 0;
#endif

		case MMCSDSECTORSIZE:

			if (!argument) return -EINVAL; /* NULL pointer: not valid */

			if(copy_to_user((uint16_t *)argument,
						&(device->phy_sectorsize),
						sizeof(uint16_t)))
			{
				return -EFAULT;
			}
			return 0;

		case MMCSDERASESECTOR:

			if (!argument) return -EINVAL; /* NULL pointer: not valid */

			if (copy_from_user(&blks, (MMCSD_Blks *)argument, sizeof (blks)))
			{
				return -EFAULT;
			}

			return(mmcsd_sector_erase(device, &blks));

		case MMCSDPROTECT:

			if (!argument) return -EINVAL; /* NULL pointer: not valid */
			if (copy_from_user(&addr,(uint32_t*)argument, sizeof(uint32_t)))
			{
				return -EFAULT;
			}
			return(mmcsd_write_protect(device,addr,1));

		case MMCSDCLEPROTECT:
			if (!argument) return -EINVAL; /* NULL pointer: not valid */
			if (copy_from_user(&addr,(uint32_t*)argument, sizeof(uint32_t)))
			{
				return -EFAULT;
			}

			return(mmcsd_write_protect(device,addr,0));

		case MMCSDSETPW:
		case MMCSDRESETPW:
		case MMCSDLOCK:
		case MMCSDUNLOCK:
		case MMCSDFORCERASE:

			pwd = (MMCSD_PWD *)malloc(2+((MMCSD_PWD *)argument)->len);
			if(pwd == NULL)
			{
				TRACE("\nmmcsd_ioctl: malloc error\n");
				return 1;
			}

			if (copy_from_user(pwd,(MMCSD_PWD *)argument, sizeof((2+((MMCSD_PWD *) argument)->len))))
			{
				return -EFAULT;
			}
			ret = mmcsd_passwd_cmd(device,pwd);

			free(pwd);

			return(ret);

		default:

			/*return blk_ioctl(inode->i_rdev, command, argument);*/
			return -1;

	}

	return 0; /* unknown command */
}

/******************************************************************************
 * Function Name: mmcsd_write_protect
 *
 * Input:
 * 		device:	 	pointer to the specified device.
 * 		addr:		the address of the indicated group.
 * 		is_set:		TRUE, if set the write protection; False, clear
 * 				the write protection.
 *
 * Value Returned:
 *		0	:	succeed
 *		1	:	fail
 *
 * Description:
 * 	This routine will set/clear the write protection bit for addressed
 * 	group.
 *
 * Modification History:
 * 	30 DEC,2001,  Yiwei  Zhao
 *****************************************************************************/

static int mmcsd_write_protect(MMCSD_Dev *device,uint32_t addr, char is_set)
{
	int ret = 0;

#ifndef MMCSD_TEST
	mmcsd_device_sem_obtain(device);
	g_selected_lock[device->port].device = device;
#endif
	if(mmcsd_is_writeprotection(device))
	{
		if(is_set)
		{
			ret = _MMCSD_SetWriteProt(device, addr);
			if(ret)
			{
				TRACE("_MMCSD_SetWriteProt error\n");
				goto write_protect_ret;
			}
		}
		else
		{
			ret = _MMCSD_ClearWriteProt(device,addr);
			if(ret)
			{
				TRACE("_MMCSD_ClearWriteProt error\n");
				goto write_protect_ret;
			}
		}
	}
	else
	{
		INFO(" The card doesn't support the Write Protection Feature\n");
		ret = MMCSD_ERROR;
		goto write_protect_ret;
	}

write_protect_ret:
#ifndef MMCSD_TEST
	mmcsd_device_sem_release(device);
#endif
	return ret;
}


/******************************************************************************
 * Function Name: mmcsd_passwd_cmd
 *
 * Input:
 * 		device: the device write to
 * 		pwd:	pointer to MMCSD_PWD structure.
 *
 * Value Returned:
 * 		0:	success.
 * 		1:	failure.
 *
 * Description:
 * 	this routine will process passwd-related commands.
 *
 * Modification History:
 * 	30 DEC,2001, Initial version.  			Yiwei Zhao
 *****************************************************************************/

static int mmcsd_passwd_cmd(	MMCSD_Dev *device,
		MMCSD_PWD *pwd)
{
	int ret,i;
	uint32_t	state,int_mask,temp32;
	uint16_t	temp16,*p16;
	int burst_len,counter;
	int total_len;		// it's different from pwd->len. the latter is the actural pwd len.
	// total_len is the total size that will be write to FIFO.


	if((device->ccc&MMCSD_CCC_LOCK) ==0)
	{
		FAILED("The card doesn't support the feature.\n");
		return MMCSD_ERROR;
	}


	total_len = (pwd->len+2) + (pwd->len%2);
	//********************************************************************
	// Get Lock!!!!!!!!!!!!!
	//********************************************************************

#ifndef MMCSD_TEST
	mmcsd_device_sem_obtain(device);
	g_selected_lock[device->port].device = device;
#endif
	//********************************************************************
	// Check Status
	//********************************************************************
	state =MMCSD_TRANSFER;
	ret = mmcsd_check_status(device,&state);
	if(ret)
	{
		TRACE("check status failed.\n");
		ret = MMCSD_ERROR;
		goto passwd_failed;
	}
	/**************************************************************
	 * Mask Interrupt
	 **************************************************************/
	TRACE("Disable Interrupt..");
	_MMCSD_Mask_interrupt(device,0x7F,&int_mask);

	/********************************************************************
	 * Get the Lock for seclected device
	 *******************************************************************/


#ifdef MMCSD_DMA_ENDIAN_ERR
	_mmcsd_16_big_2_little((uint8_t*)pwd, total_len);
#endif
	/********************************************************************
	 * Define the block length
	 *******************************************************************/

	ret = _MMCSD_SetBlockLen(device, 2+pwd->len,1);
	if(ret)
	{
		FAILED("SetBlockLen error");
		ret = MMCSD_ERROR;
		goto passwd_failed;
	}

	if(device->bit_width == 1)
	{
		ret = _MMCSD_SetBusWidth(device, device->card_rca,2);
		if(ret)
		{
			FAILED("Set Bus Width failed.");
			ret = MMCSD_ERROR;
			goto passwd_failed;
		}
	}
	else
	{
		ret = _MMCSD_SetBusWidth(device, device->card_rca,0);
		if(ret)
		{
			FAILED("Set Bus Width failed.");
			ret = MMCSD_ERROR;
			goto passwd_failed;
		}
	}
	//	while(_MMCSD_IsRun())
	_MMCSD_StopClk(0); 	//Ricky

	/********************************************************************
	 * Process Passwd command
	 *******************************************************************/
	counter = 0;
	p16 = (uint16_t *)pwd;

	TRACE("Begin Set Command..\n");
	if(device->bit_width == 0)
	{
		_MMCSD_SetCmd(device,MMCSD_CMD42,0,MMCSDB_DATEN|MMCSDB_R1|MMCSDB_WRRD);
		burst_len = 8;
	}
	else
	{
		_MMCSD_SetCmd(device,MMCSD_CMD42,0,MMCSD_BUS_4BIT|MMCSDB_DATEN|MMCSDB_R1|MMCSDB_WRRD);
		burst_len = 32;
	}
	device->cur_cmd = MMCSD_CMD42;
	_MMCSD_StartClk(0);
	/*	if((ret=_MMCSD_WaitUntil(MMCSD_CMD_DONE))!=0)
		{
		FAILED("WaitCMDDone ret 0x%lx",ret);
		goto passwd_failed;
		}
	 */
	ret = MMCSD_OK;
	while(ret == MMCSD_OK)
	{
		uint32_t status;

		status =mmcsd_reg_rd(device, MX1_MMCSD_STATUS_o);
		if((status&0x1047)==0)
		{
			TRACE("MMCSD_STATUS is 0x%0X\n",status);
			continue;
		}

		//		TRACE("MMCSD_STATUS = 0x%x\n",status);

		if(status&0x0007)
		{
			TRACE("7)ERROR!--, MMCSD_STATUS = %lx,counter = %d\n",(unsigned long)status,counter);
			ret = MMCSD_ERROR;
			goto passwd_failed;
		}
		else if((status&0x0040)!=0) // Buffer is empty
		{
			//			TRACE("FIFO is empty!\n");
			for(i = 0;(i<burst_len)&&(counter<total_len); i++)
			{
				temp16 = (*p16);
				temp32 = temp16;
				mmcsd_reg_wr(device, MX1_MMCSD_BUFFER_ACCESS_o, temp32);
				p16++;
				counter += 2;
			}
		}
		if(counter >= total_len)
		{
			break;
		}
	}

	/*******************************************************
	 * Write Done
	 *******************************************************/
	if(ret==MMCSD_OK)
	{
		counter = 0xFFFF;
		while(counter--)
		{
			uint32_t status;

			status =mmcsd_reg_rd(device, MX1_MMCSD_STATUS_o);
			if(status&0x0007)
			{
				FAILED("8)ERROR! status = 0x%lx",(unsigned long)status);
				ret = MMCSD_ERROR;
				break;
			}
			else if((status&0x1000)!=0)
			{
				TRACE("9)Write Done! MMCSD_STATUS = 0x%lx\n",(unsigned long)status);
				ret = MMCSD_OK;
				break;
			}
			TRACE("10)MMCSD_STATUS = %lx\n",(unsigned long)status);
		}
	}

passwd_failed:
	//	_MMCSD_StopClk();	//Ricky

	TRACE("Enable Intr...\n");
	_MMCSD_Restore_Mask(device,int_mask);
	state =MMCSD_TRANSFER;
	mmcsd_check_status(device,&state);
#ifndef MMCSD_TEST
	mmcsd_device_sem_release(device);
#endif

	return ret;
}

/******************************************************************************
 * Function Name: mmcsd_is_writeprotection
 *
 * Input:
 * 		device: the device write to
 *
 * Value Returned:
 * 		1:	the device support protection mechanism.
 *
 * Description:
 * 	this routine test whether the device support protection mechanism.
 *
 * Modification History:
 * 	30 DEC,2001, Initial version.  			Yiwei Zhao
 *****************************************************************************/
static int mmcsd_is_writeprotection(MMCSD_Dev *device)
{
	if((device->ccc&MMCSD_CCC_WRPT)!=0)
	{
		return 1;
	}
	else
	{
		return 0;
	}
}


int mmcsd_getMBR(MMCSD_Dev * device)
{
	int ret=MMCSD_OK,i;
	uint8_t * buf = NULL;

	buf = malloc(MMCSD_FS_SECTOR_SIZE*sizeof(uint8_t));
	if(!buf)
	{
		FAILED("malloc failed");
		return MMCSD_ERROR;
	}

	for(i = 0; i < 512; i++)
	{
		*(buf+i)=0xAA;
	}
	ret = mmcsd_read_device(device,
			buf,
			0,
			1);
	if(ret)
	{
		FAILED("mmcsd_read_device() error!");
		ret = MMCSD_ERROR;
		goto getMBR_error;
	}


	device->lba_sec_offset = (*(buf+0x1C6))    |
		(*(buf+0x1C7))<<8 |
		(*(buf+0x1C8))<<16|
		(*(buf+0x1C9))<<24;
	TRACE("lba_sec_offet is 0x%lx\n",(unsigned long)device->lba_sec_offset);
	TRACE("sect_num is 0x%lx\n",(unsigned long)device->sect_num);

	if(device->lba_sec_offset > device->sect_num)
	{
		// PLAM -- if MBR is indeed a PBR, the sector offset to this partition is zero!
		if(*buf==0xeb && *(buf+1)==0x3c && *(buf+2)==0x90) // starting bytes of a PBR
		{
			TRACE("PBR in place of MBR, lba_sec_offset is now set to 0\n");
			device->lba_sec_offset = 0;
		}
		else
		{
			// end PLAM
			INFO("LBA is Wrong (%ld)\n",(long int)(device->lba_sec_offset));
			ret = MMCSD_ERROR;
			// PLAM
		}
		// end PLAM
	}

getMBR_error:

	free(buf);
	return ret;

}
void _mmcsd_16_big_2_little(uint8_t *buf, uint32_t size)
{
	uint8_t  *p;
	uint16_t temp16;
	uint32_t j;

	p = buf;

	for(j = 0; j< size/2; j++)
	{
		temp16 = (*(uint16_t *)p);


		*(uint8_t *)p = (temp16&0xFF00)>>8;
		(uint8_t *)p++;
		*(uint8_t *)p = (temp16&0x00FF);
		(uint8_t*)p++;
	}
}
void _mmcsd_cp_16_little_2_big(uint8_t *src,uint8_t *target, int size)
{
	int i;
	uint16_t tmp16,*p16;

	p16 = (uint16_t *)target;
	for(i = 0;i < size/2; i++)
	{
		tmp16 = (*src<<8);
		src++;
		tmp16 = tmp16|(*src);
		src++;
		*p16 = tmp16;
		p16++;

	}
}
void _mmcsd_print_page(uint8_t *cur_buf, uint32_t size)
{
	int j;

	for(j = 0; j < size; j++)
	{
		if(j%16 == 0) printf("\n 0x%04X\t ---",j);
		printf(" %02x ", (uint8_t)(*(cur_buf+j)));
	}
	printf("\n");
}

uint8_t * test_buf[MMCSD_TEST_PAGE_NUM];
void _mmcsd_test_reading_pages(MMCSD_Dev * device)
{
	uint8_t i;
	uint32_t ret;

	for( i = 0;i < MMCSD_TEST_PAGE_NUM; i++)
	{
		test_buf[i] = malloc(MMCSD_FS_SECTOR_SIZE*sizeof(uint8_t));
	}
	//	buf2=buf3=buf4=buf1;


	for( i = 0;i < MMCSD_TEST_PAGE_NUM; i++)
	{
		memset(test_buf[i],0xAA,MMCSD_FS_SECTOR_SIZE);
		ret = mmcsd_read_device(device,
				test_buf[i],
				TESTING_PAGE+i,
				1);
		TRACE("Sector %d\n",TESTING_PAGE+i);
		_mmcsd_print_page(test_buf[i],512);
	}

	for( i = 0;i < MMCSD_TEST_PAGE_NUM; i++)
	{
		free(test_buf[i]);
	}
}

void mmcsd_delay(uint32_t delay)
{
	while(delay--);
}


#ifdef MMCSD_WITH_PM

static int handle_pm_event(struct pm_dev *dev, pm_request_t rqst, void *data)
{
	MMCSD_Dev *device1;
#ifdef CONFIG_ARCH_MX2ADS
	MMCSD_Dev *device2;

	device2 = &g_mmcsd_devices[1];
#endif
	device1 = &g_mmcsd_devices[0];

	switch(rqst)
	{
		case PM_SUSPEND:
			INFO("MMC will comes into Suspend state");
			mmcsd_device_lock(device1);
			mmcsd_device_sem_obtain(device1);
#ifdef CONFIG_ARCH_MX2ADS
			mmcsd_device_lock(device2);
			mmcsd_device_sem_obtain(device2);
			//pin pe18~23,18,19,20,22 up, 21-dat3,23-clk
			_reg_PTE_DDIR &= ~0x00fc0000;
			_reg_PTE_GIUS |= 0xff03ffff;
			_reg_PTE_PUEN |= 0x00fc0000;

			//add for SD2 pb4~pb9 pb7-dat3,pb9-clk
			_reg_PTB_DDIR &= ~0x000003f0;
			_reg_PTB_GIUS |= 0x000003f0;
			_reg_PTB_PUEN |= 0x000003f0;
			_MMCSD_StopClk(device1);
			_MMCSD_StopClk(device2);


#endif
			//pb8~13, 11-dat3,12-clk,13-cmd
#ifdef CONFIG_ARCH_MX1ADS
			/*
			 _reg_PTB_DDIR &= ~0x00003F00;		// Enable SDHC's port to Output
			 _reg_PTB_GIUS |= 0x00003F00;
			 _reg_PTB_PUEN |= 0x00003F00;
			 */
			_MMCSD_StopClk(device1);
#endif
			break;
		case PM_RESUME:
			INFO("MMC will be resumed.\n");

			mmcsd_device_sem_release(device1);
			mmcsd_device_unlock(device1);
#ifdef CONFIG_ARCH_MX2ADS
			mmcsd_device_sem_release(device2);
			mmcsd_device_unlock(device2);
			if(device2->card_type != MMCSD_NO_CARD)
			{

				_MMCSD_InitSDHC(device2);
				_MMCSD_64M_MMC_PinConfig();
			}
#endif
			if(device1->card_type != MMCSD_NO_CARD)
			{
				_MMCSD_InitSDHC(device1);
				_MMCSD_64M_MMC_PinConfig();
			}
			_MMCSD_SetClk(device1,device1->max_dtr);
#ifdef CONFIG_ARCH_MX2ADS
			_MMCSD_SetClk(device2,device2->max_dtr);
#endif

			break;
		default:
			break;
	}
	return 0;
}

#endif /*MMCSD_WITH_PM*/
